package com.testing;

import static org.junit.jupiter.api.Assertions.*;

import org.junit.jupiter.api.Test;

import com.inspect.Inspector;
import com.nodes.InfoNode;
import com.nodes.Node;

class NodeTesting {
	
	public static class TestStubA {
		int value = 4;
	}
	
	public static class TestStubB extends TestStubA {
		
	}
	
	interface ITest {
		public void wow();
	}
	
	public static class TestStubC implements ITest {
		@Override
		public void wow() {
			// TODO Auto-generated method stub
			
		}
	}
	
	public static class TestStubD {
		int value = 4;
		
		private static int hello() {
			return 2;
		}
		
		protected final synchronized int hello2() {
			return 4;
		}
		
	}
	
	public static class TestStubE {
		private int value1 = 0;
		protected String value2 = "what";
		public float value3 = 0.3f;
	}
	
	public static class TestStubEABC {
		protected String value2;
		private int value1;
		public float value3;
		
	}
	
	public static class TestStubF {
		private int[] hello = { 1, 2, 3 };
		
	}

	public static class TestStubG {
		
		public TestStubG(int one) throws IllegalAccessException {
			
		}
		
		private void why(int two) throws ExceptionInInitializerError {
		}
	}
	
	public static class TestStubH {
		public TestStubA wow = new TestStubA();
		public TestStubA[] wowa = new TestStubA[] { new TestStubA() };
	
	}
	
	@Test
	void testTestStubA() throws IllegalArgumentException, IllegalAccessException, NoSuchFieldException {
		TestStubA stub = new TestStubA();
		Inspector inspector = new Inspector();
		
		Node root = Node.root();
		inspector.properInspection(root, stub, stub.getClass(), false, 0);
		

		// Verify the class details are correct.
		assertEquals(root.lookupKey("Class").get(), "com.testing.NodeTesting$TestStubA");
		
		// Verify the fields are correct.
		Node fieldNode = root.lookupEntrypoint("Fields").get().getChild(0).get();
		assertEquals(fieldNode.lookupKey("Name").get(), "value");
		assertEquals(fieldNode.lookupKey("Type").get(), "int");
		assertEquals(fieldNode.lookupKey("Value").get(), "4");
	
		// Verify the superclass is correct.
		Node superclassNode = root.lookupEntrypoint("Superclass").get();
		assertEquals(superclassNode.lookupKey("Class").get(), "java.lang.Object");
		
		// Verify the constructor is correct.
		Node constructors = root.lookupEntrypoint("Constructors").get().getChild(0).get();
		assertEquals(constructors.lookupKey("Name").get(), "com.testing.NodeTesting$TestStubA");
		assertEquals(constructors.lookupKeyList("Modifiers").get()[0], "public");
	}

	
	
	@Test
	void verifySuperclassesExploredProperly() throws IllegalArgumentException, IllegalAccessException, NoSuchFieldException {
		TestStubA stub = new TestStubB();
		Inspector inspector = new Inspector();
		
		Node root = Node.root();
		inspector.properInspection(root, stub, stub.getClass(), false, 0);
		

		// Verify the class details are correct.
		assertEquals(root.lookupKey("Class").get(), "com.testing.NodeTesting$TestStubB");
		assertEquals(root.lookupEntrypoint("Superclass").get().lookupKey("Class").get(), "com.testing.NodeTesting$TestStubA");
		assertEquals(root.lookupEntrypoint("Superclass").get().lookupEntrypoint("Superclass").get().lookupKey("Class").get(), "java.lang.Object");

	}
	
	@Test
	void verifyInterfacesExplored() throws IllegalArgumentException, IllegalAccessException, NoSuchFieldException {
		TestStubC stub = new TestStubC();
		Inspector inspector = new Inspector();
		
		Node root = Node.root();
		inspector.properInspection(root, stub, stub.getClass(), false, 0);
		

		// Verify the interface details are correct.
		Node i = root.lookupEntrypoint("Interfaces").get().getChild(0).get();
		assertEquals(i.lookupKey("Class").get(), "com.testing.NodeTesting$ITest");
		assertEquals(i.lookupEntrypoint("Methods").get().getChild(0).get().lookupKey("Name").get(), "wow");
		String[] modifiers = i.lookupEntrypoint("Methods").get().getChild(0).get().lookupKeyList("Modifiers").get();
		assertEquals(modifiers[0], "public");
		assertEquals(modifiers[1], "abstract");
		
		// Verify methods
		Node m = root.lookupEntrypoint("Methods").get().getChild(0).get();
		assertEquals(m.lookupKey("Name").get(), "wow");
		assertEquals(m.lookupKey("Return Type").get(), "void");
		assertEquals(m.lookupKeyList("Modifiers").get()[0], "public");
	}
	
	
	@Test
	void verifyMethodsExplored() throws IllegalArgumentException, IllegalAccessException, NoSuchFieldException {
		TestStubD stub = new TestStubD();
		Inspector inspector = new Inspector();
		
		Node root = Node.root();
		inspector.properInspection(root, stub, stub.getClass(), false, 0);
		
		// Methods
		Node m = root.lookupEntrypoint("Methods").get();
		assertEquals(m.getChild(0).get().lookupKey("Name").get(), "hello");
		assertEquals(m.getChild(0).get().lookupKey("Return Type").get(), "int");
		assertArrayEquals(m.getChild(0).get().lookupKeyList("Modifiers").get(), new String[] { "private", "static" });
		
		assertEquals(m.getChild(1).get().lookupKey("Name").get(), "hello2");
		assertEquals(m.getChild(1).get().lookupKey("Return Type").get(), "int");
		assertArrayEquals(m.getChild(1).get().lookupKeyList("Modifiers").get(), new String[] { "protected", "final", "synchronized" });

	}
	
	@Test
	void verifyFieldsExplored() throws IllegalArgumentException, IllegalAccessException, NoSuchFieldException {
		TestStubE stub = new TestStubE();
		Inspector inspector = new Inspector();
		
		Node root = Node.root();
		inspector.properInspection(root, stub, stub.getClass(), false, 0);
		
		// Fields
		Node f = root.lookupEntrypoint("Fields").get();
		assertEquals(f.getChild(0).get().lookupKey("Name").get(), "value1");
		assertEquals(f.getChild(0).get().lookupKey("Type").get(), "int");
		assertArrayEquals(f.getChild(0).get().lookupKeyList("Modifiers").get(), new String[] { "private" });
		assertEquals(f.getChild(0).get().lookupKey("Value").get(), "0");
		
		assertEquals(f.getChild(1).get().lookupKey("Name").get(), "value2");
		assertEquals(f.getChild(1).get().lookupKey("Type").get(), "java.lang.String");
		assertArrayEquals(f.getChild(1).get().lookupKeyList("Modifiers").get(), new String[] { "protected" });
		assertEquals(f.getChild(1).get().lookupKey("Value").get(), "what");
		
		assertEquals(f.getChild(2).get().lookupKey("Name").get(), "value3");
		assertEquals(f.getChild(2).get().lookupKey("Type").get(), "float");
		assertArrayEquals(f.getChild(2).get().lookupKeyList("Modifiers").get(), new String[] { "public" });
		assertEquals(f.getChild(2).get().lookupKey("Value").get(), "0.3");
		
	}
	
	@Test
	void verifyFieldsExploredAlphabetical() throws IllegalArgumentException, IllegalAccessException, NoSuchFieldException {
		TestStubEABC stub = new TestStubEABC();
		Inspector inspector = new Inspector();
		
		Node root = Node.root();
		inspector.properInspection(root, stub, stub.getClass(), false, 0);
		
		// Fields
		Node f = root.lookupEntrypoint("Fields").get();
		assertEquals(f.getChild(0).get().lookupKey("Name").get(), "value1");
		assertEquals(f.getChild(1).get().lookupKey("Name").get(), "value2");		
		assertEquals(f.getChild(2).get().lookupKey("Name").get(), "value3");
	}
	
	
	@Test
	void verifyArrayFields() throws IllegalArgumentException, IllegalAccessException, NoSuchFieldException {
		TestStubF stub = new TestStubF();
		Inspector inspector = new Inspector();
		
		Node root = Node.root();
		inspector.properInspection(root, stub, stub.getClass(), false, 0);
		
		// Fields
		Node f = root.lookupEntrypoint("Fields").get();
		assertArrayEquals(f.getChild(0).get().lookupKeyList("Values").get(), new String[] { "1", "2", "3" });
	}
	
	@Test
	void verifyExceptionExtraction() throws IllegalArgumentException, IllegalAccessException, NoSuchFieldException {
		TestStubG stub = new TestStubG(0);
		Inspector inspector = new Inspector();
		
		Node root = Node.root();
		inspector.properInspection(root, stub, stub.getClass(), false, 0);
		
		// Method
		Node f = root.lookupEntrypoint("Methods").get().getChild(0).get();
		assertArrayEquals(f.lookupKeyList("Exceptions").get(), new String[] { "java.lang.ExceptionInInitializerError" } );
	
		// Construcor
		Node c = root.lookupEntrypoint("Constructors").get().getChild(0).get();
		assertArrayEquals(c.lookupKeyList("Exceptions").get(), new String[] { "java.lang.IllegalAccessException" } );
	
	}
	
	@Test
	void verifyParameterExtraction() throws IllegalArgumentException, IllegalAccessException, NoSuchFieldException {
		TestStubG stub = new TestStubG(0);
		Inspector inspector = new Inspector();
		
		Node root = Node.root();
		inspector.properInspection(root, stub, stub.getClass(), false, 0);
		
		// Method
		Node f = root.lookupEntrypoint("Methods").get().getChild(0).get();
		assertArrayEquals(f.lookupKeyList("Parameter Types").get(), new String[] { "int" } );
	
		// Construcor
		Node c = root.lookupEntrypoint("Constructors").get().getChild(0).get();
		assertArrayEquals(c.lookupKeyList("Parameter Types").get(), new String[] { "int" } );
	}
	
	@Test
	void verifyRecursionOnField() throws IllegalArgumentException, IllegalAccessException, NoSuchFieldException {
		TestStubH stub = new TestStubH();
		Inspector inspector = new Inspector();
		
		Node root = Node.root();
		inspector.properInspection(root, stub, stub.getClass(), true, 0);
		
		// Method
		Node f = root.lookupEntrypoint("Fields").get().getChild(0).get().lookupEntrypoint("Value").get();
		assertEquals(f.lookupKey("Class").get(), "com.testing.NodeTesting$TestStubA");
		
		
		Node c = root.lookupEntrypoint("Fields").get().getChild(1).get().lookupEntrypoint("Values").get();
		assertEquals(c.getChild(0).get().lookupKey("Class").get(), "com.testing.NodeTesting$TestStubA");
	}
	
	@Test
	void verifyNonRecursiveObjectFields() throws IllegalArgumentException, IllegalAccessException, NoSuchFieldException {
		TestStubH stub = new TestStubH();
		Inspector inspector = new Inspector();
		
		Node root = Node.root();
		inspector.properInspection(root, stub, stub.getClass(), false, 0);
		
		// Method
		Node f = root.lookupEntrypoint("Fields").get().getChild(0).get();
		assertEquals(f.lookupKey("Value").get(), stub.wow.toString());

		String[] c = root.lookupEntrypoint("Fields").get().getChild(1).get().lookupKeyList("Values").get();
		assertArrayEquals(c, new String[] { stub.wowa[0].toString() } );
	}
}
